#include <iostream> 
#include "Run_Parameter.H"
#include "MathTools.H"
#include "Message.H"
#include "Exception.H"
#include "Random.H"
#include "Data_Reader.H"
#include "Data_Collector.H"
#include "MyStrStream.H"
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>
#include <pwd.h>

using namespace ATOOLS;

bool ATOOLS::Run_Parameter::s_initialized=false;
std::map<std::string,std::string> ATOOLS::Run_Parameter::s_variables;
Run_Parameter ATOOLS::rpa;

Run_Parameter::Run_Parameter() 
{
  AnalyseEnvironment();
  gen.m_output    = gen.m_analysis   = 0;
  gen.m_nevents   = 0;
  gen.m_cutscheme = 0;
  gen.m_ecms      = gen.m_ycut       = gen.m_accu        = 0.;
  gen.m_delta_r   = 1.;
  gen.m_beam1     = gen.m_beam2      = Flavour(kf::none);
  gen.m_rpa_id    = gen.m_flavour_id = std::string("");
#ifndef USING__ATOOLS_only
  gen.p_model     = NULL;
#endif
  gen.m_ndicedevents = 0;
  gen.m_batchmode = 1;
  gen.SetTimeOut(3600);
  gen.SetWAnaScale(1.);
  gen.m_spincorrelations = 0;
} 

std::ostream &ATOOLS::operator<<(std::ostream &str,const Run_Parameter &rp)
{ 
  return str<<"("<<&rp<<"): {\n}"; 
}

void Run_Parameter::AnalyseEnvironment() 
{
  if (s_initialized) return;
#ifdef __GNUC__
#if __GNUC__ == 2 && __GNUC_MINOR__ == 96
#error Sherpa was not designed for gcc 2.96
#endif
#endif
  std::string gccversion;
  system("gcc -dumpversion > sherpa_gcc_test");
  std::ifstream *test = new std::ifstream("sherpa_gcc_test");
  if (*test) (*test)>>gccversion;
  delete test;
  system("if test -f sherpa_gcc_test; then rm sherpa_gcc_test; fi");
  if (gccversion.find("2.96")!=std::string::npos) {
    THROW(fatal_error,"Sherpa must not be run on gcc version 2.96 !");
  }
  char *var=NULL;
  s_variables["PATH"]=std::string(((var=getenv("PATH"))==NULL?"":var));
  s_variables["SHERPASYS"]=std::string(((var=getenv("SHERPASYS"))==NULL?"":var));
  s_variables["SHERPA_PDF_PATH"]=std::string(((var=getenv("SHERPA_PDF_PATH"))==NULL?"":var));
  s_variables["SHERPA_CPP_PATH"]=std::string(((var=getenv("SHERPA_CPP_PATH"))==NULL?"":var));
  s_variables["SHERPA_LIB_PATH"]=std::string(((var=getenv("SHERPA_LIB_PATH"))==NULL?"":var));
  s_variables["LD_LIBRARY_PATH"]=std::string(((var=getenv("LD_LIBRARY_PATH"))==NULL?"":var));
  s_variables["SHERPA_BIN_PATH"]=SHERPA_BINARY_PATH;
  s_variables["SHERPA_RUN_PATH"]=getenv("PWD");
  s_initialized=true;
}

void Run_Parameter::Init(std::string path,std::string file,int argc,char* argv[])
{
  gen.m_timer.Start();
  struct passwd* user_info = getpwuid(getuid());
  if (!user_info) gen.m_username="<unknown user>";
  else gen.m_username=user_info->pw_gecos;
  m_path = path;
  Data_Read dr(m_path+file);
  gen.m_output = dr.GetValue<int>("OUTPUT",0);
  std::string logfile=dr.GetValue<std::string>("LOG_FILE",std::string(""));
  msg.Init(gen.m_output,logfile);
  if (msg.Level()>0) 
    msg.Out()<<"Welcome to Sherpa, "<<gen.m_username
	     <<". Initialization of framework underway."<<std::endl;
  // make path nice
  if (path.length()>0) {
    if (path[0]!='/') path=std::string(getenv("PWD"))+"/"+path;
    while (path.length()>0 && 
	   path[path.length()-1]=='/' || path[path.length()-1]=='.') 
      path=path.substr(0,path.length()-1);
  }
  // set pdf path
  std::string pdfpath=dr.GetValue<std::string>("SHERPA_PDF_PATH",std::string(""));
  if (pdfpath.length()>0 && pdfpath[0]=='/') s_variables["SHERPA_PDF_PATH"]=pdfpath;
  else if (s_variables["SHERPA_PDF_PATH"].length()==0) 
    s_variables["SHERPA_PDF_PATH"]=SHERPA_SHARE_PATH;
  s_variables["SHERPA_SHARE_PATH"]=SHERPA_SHARE_PATH;
  // set cpp path
  std::string cpppath=dr.GetValue<std::string>("SHERPA_CPP_PATH",std::string(""));
  if (cpppath.length()>0 && cpppath[0]=='/') s_variables["SHERPA_CPP_PATH"]=cpppath;
  else if (path!=getenv("PWD")) s_variables["SHERPA_CPP_PATH"]=path;
  else if (s_variables["SHERPA_CPP_PATH"].length()==0) 
    s_variables["SHERPA_CPP_PATH"]=s_variables["SHERPA_RUN_PATH"];
  // set lib path
  std::string libpath=dr.GetValue<std::string>("SHERPA_LIB_PATH",std::string(""));
  if (libpath.length()>0 && libpath[0]=='/') s_variables["SHERPA_LIB_PATH"]=libpath;
  else if (s_variables["SHERPA_LIB_PATH"].length()==0) 
    s_variables["SHERPA_LIB_PATH"]=s_variables["SHERPA_CPP_PATH"]
      +std::string("/Process/lib");
  s_variables["SHERPA_INC_PATH"]=SHERPA_INCLUDE_PATH;
  msg_Tracking()<<"Run_Parameter::Init(..): Paths are {\n"
		<<"   SHERPA_BIN_PATH = "<<s_variables["SHERPA_BIN_PATH"]<<"\n"
		<<"   SHERPA_PDF_PATH = "<<s_variables["SHERPA_PDF_PATH"]<<"\n"
		<<"   SHERPA_CPP_PATH = "<<s_variables["SHERPA_CPP_PATH"]<<"\n"
		<<"   SHERPA_LIB_PATH = "<<s_variables["SHERPA_LIB_PATH"]<<"\n"
		<<"   SHERPA_INC_PATH = "<<s_variables["SHERPA_INC_PATH"]<<"\n"
		<<"}"<<std::endl;
#ifndef __sgi
  setenv("LD_LIBRARY_PATH",(s_variables["LD_LIBRARY_PATH"]+std::string(":")+
			    s_variables["SHERPA_LIB_PATH"]).c_str(),1);
#endif
  gen.m_analysis           = dr.GetValue<int>("ANALYSIS",0);
  gen.m_nevents            = dr.GetValue<long>("EVENTS",100);
  // read only if defined (no error message if not defined)

  Data_Reader dreader(" ",";","!","=");
  dreader.AddWordSeparator("\t");
  dreader.SetInputFile(m_path+file);
  std::vector<long int> seeds;
  gen.m_seed2 = -1;
  if (dreader.VectorFromFile(seeds,"RANDOM_SEED")) {
    gen.m_seed = seeds[0];
    // if 2nd seed is given, store it
    if (seeds.size() == 2) { gen.m_seed2 = seeds[1]; } 
  } else gen.m_seed=1234;

  gen.m_timeout            = dr.GetValue<double>("TIMEOUT");
  if (gen.m_timeout<0.) gen.m_timeout=0.;
  rpa.gen.m_timer.Start();
  gen.m_batchmode          = dr.GetValue<int>("BATCH_MODE");
  if (gen.m_batchmode==NotDefined<int>()) gen.m_batchmode=1;
  int stacktrace           = dr.GetValue<int>("STACK_TRACE");
  if (stacktrace==NotDefined<int>()) stacktrace=1;
  exh->SetStackTrace(stacktrace);
  double ycut=dr.GetValue<double>("YCUT");
  if (ycut!=NotDefined<double>()) gen.m_ycut=ycut;
  gen.m_accu               = dr.GetValue<double>("Num. Accuracy",1.e-10);
  //gen.m_runtime            = dr.GetValue<std::string>("Runtime"); // Time
  Switch::code color=dr.GetValue<Switch::code>("PRETTY_PRINT",Switch::On);
  if (color==Switch::On) {
    termios testos;
    if (tcgetattr(STDOUT_FILENO,&testos)==0) msg.SetModifiable(true);
  }
  gen.m_rpa_id = dr.GenerateKey();
  if (gen.m_seed2!=-1) { ran.SetSeed(gen.m_seed, gen.m_seed2); }
                  else { ran.SetSeed(gen.m_seed); }
  gen.m_ren_scale_fac=1.0;
  gen.m_fac_scale_fac=1.0;
  gen.m_pdf_scalefac=dr.GetValue<double>("PDF_SCALE_FACTOR",1.0);
  gen.m_as_scalefac=dr.GetValue<double>("AS_SCALE_FACTOR",1.0);
  if(gen.m_pdf_scalefac!=1.0 || gen.m_as_scalefac!=1.0) {
    PRINT_INFO("Attention: Operator approach for scale variations activated.  mu_f="
	       <<gen.m_pdf_scalefac<<";  mu_r="<<gen.m_as_scalefac);
  }
}

Run_Parameter::~Run_Parameter() 
{ 
  if (msg.Level()>=1) gen.m_timer.PrintTime();
}

bool Run_Parameter::Gen::CheckTime(const double limit)
{ 
  if (limit==0.) {
    if (m_timeout>0.) return m_timer.UserTime()<m_timeout;
  }
  else {
    return m_timer.UserTime()<limit;
  }
  return false;
}

void  Run_Parameter::Gen::SetEcms(double _ecms)     { 
  m_ecms    = _ecms;
  Data_Collector::AddData("ECMS",new Blob_Data<double>(m_ecms));
}
void  Run_Parameter::Gen::SetPBeam(short unsigned int i,Vec4D pbeam) { 
  m_pbeam[i]=pbeam;
  //  Data_Collector::AddData("PBEAM"+ToString(i),new Blob_Data<Vec4D>(m_pbeam[i]));
}
void  Run_Parameter::Gen::SetBeam1(const Flavour b) { 
  m_beam1  = b;   
  Data_Collector::AddData("BEAM1",new Blob_Data<std::string>(m_beam1.TexName()));
}
void  Run_Parameter::Gen::SetBeam2(const Flavour b) { 
  m_beam2  = b;   
  Data_Collector::AddData("BEAM2",new Blob_Data<std::string>(m_beam2.TexName()));
}

std::string Run_Parameter::Gen::Variable(const std::string &key,const std::string &def) 
{ 
  return s_variables.find(key)!=s_variables.end()?s_variables[key]:def; 
}
